# -*- perl -*-
#
# Package to interface with LDOCE3 machine readable dictionary 
#
# Run 'perldoc' on this file to produce documentation
#
# This module is free software; you can redistribute it and/or modify
# it under the same terms as Perl itself.
#
#
# Created for CS 8761 - Final Project
# Beta code for an interface to the LDOCE MRD
#
# Team: Alianza Lima
# Team Members:
# 	Abou-Rjeili, Amine
#	Goopy, Suchitra	
#	Purandare, Amruta	
#	Storie, Sam
#	


####### manual page & loadIndex ##########

package LDOCE::Interface;

use strict;
use Exporter;

##############################
# Environment/Initialization #
##############################

BEGIN {
    use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
    # List of classes from which we are inheriting methods
    @ISA = qw(Exporter);
    # Automatically loads these function names to be used without qualification
    @EXPORT = qw();
    # Allows these functions to be used without qualification
    @EXPORT_OK = qw();
    $VERSION = do { my @r=(q$Revision: 0.01 $=~/\d+/g); sprintf "%d."."%02d"x$#r,@r };
}

#############################
# Private Package Variables #
#############################

my @LETTERS = ('A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z');

my %alphabet;
my %subjectCodes; # The data sorted by subject codes
my $subjectCodesLoaded = 0;
my %posCodes;
my %wsMap; # Maps words to all the words that have a defn that contains the word (key)
my $posCodesLoaded = 0;
my @dictFiles;

# Similar to what is done in WordNet::QueryData because, well, it's a nice
#	way to do this :)
#
my $ldoceHome = defined($ENV{"LDOCEHOME"}) ? $ENV{"LDOCEHOME"} : "/home/cs/tpederse/CS8761/LDOCE3/A-Z/sgml";



END { } # module clean-up code here (global destructor)

###############
# Subroutines #
###############

sub _initialize
{
	my $self = shift;
	warn "Reading files for dictionary.\n" if ( $self->{verbose} );
	
	# Now get a listing of all the files in the LDOCE directory
	#
	opendir LDOCEHOME, $ldoceHome || die "Unable to read $ldoceHome\nReason: $!\n\n";
	@dictFiles = grep /\.SGM$/i, readdir LDOCEHOME;	

	# If the user requested to load all the dict at startup, then do so!!
	# 
	if( $self->{initAll} == 1 ){
	    warn "Initializing dictionary databases at startup\n" if $self->{verbose};
	    if( $posCodesLoaded == 0 ){
			_initPosCodes($self);
			$posCodesLoaded = 1;
	    }
	    if( $subjectCodesLoaded == 0 ){
			_initSubjectCodes($self);
			$subjectCodesLoaded = 1;
	    }
	}
}


# Argument listing: directory, initAll,  verbose
sub new
{
	# Similar to Wordnet we're providing the option of specifying the
	# 	location of the various SGML files when the class is instantiated
	# 	
    # The first argument is the class itself
    my $class = shift;
    
    my $self = {};

	# Make $self a reference to this class
	#
    bless $self, $class;
	
	# The second is the optional directory
    $self->{dir} = shift if (defined(@_ > 0));
    
    # The third is start initialization flag
    $self->{initAll} = @_ ? shift : 0;
    
 	# The fourth is the option of providing verbose output
	$self->{verbose} = @_ ? shift : 0;

    warn "Dir = ", $self->{dir}, "\n" if ($self->{verbose});
    warn "Verbose = ", $self->{verbose}, "\n" if ($self->{verbose});
    warn "InitAll = ", $self->{initAll}, "\n" if ($self->{verbose});

	
	$self->_initialize ();
    return $self;
}


# This function could be useful to initialize everything right away
#
sub InitializeAllLetters
{
	my $self = shift;
	my $letter;

	# Just cycle over every letter and initialize it
	#
	foreach $letter (@LETTERS)
	{
		InitializeLetter($self, $letter);
	}
}


# This will blindly initialize the values for a letter. It does not care
#	if it has already been done before
#
sub InitializeLetter
{
    my ($self, $letter) = @_;
    my $file;
    my $letterFile;
    my $sgmlData;
    my $entry;
    my $FILE;
    
    # We need to initialize a new letter by reading the file and 
    # 	creating a new hash for the entries
    #
    # Find the file that matches this letter
    foreach $file (@dictFiles)
    {
	if( $file =~ /^$letter/i )
	{
	    $letterFile = "$ldoceHome/$file";
	}
    }
    
    # Now if $letterFile is still undefined then there was a problem
    #	with this word (ie, it's a number of something)
    if( defined( $letterFile ) )
    {
	warn "Initializing the data from $letterFile.\n" if ($self->{verbose});
	
	# Now open the associated letter file
	open FILE, $letterFile || warn "Couldn't open file: $FILE\nReason: $!\n";
	local $/;
	$sgmlData = <FILE>;
	close FILE;
	
	# Now break the file into it's different <Entry>...</Entry> strings
	while( $sgmlData =~ m/(<Entry>.*?<\/Entry>)/sg )
	{
	    # Get the actual word in this entry and hash on it
	    $entry = $1;
	    
	    # Remove any line breaks
	    #
	    # DOUBLE CHECK IF THIS IS GOOD TO DO
	    $entry =~ s/[\r\n]//g;
	    
	    if( $entry =~ /<HWD>(.*?)<\/HWD>/ )
	    {
		if( !defined $alphabet{$letter}{$1} )
		{
		    $alphabet{$letter}{$1} = $entry;
		}
		else
		{
		    $alphabet{$letter}{$1} .= "\n$entry";
		}
	    }
	}
    }
}	

sub _initSubjectCodes( $ )
{
    my( $self ) = @_;
    my( $word, $val, $letter );
    
    # It is not defined so we need to load it and the whole hash while we are at it
    # Go through all the alphabet and load it
    warn "START: Initializing the subject codes from the dictionary ...\n" if $self->{verbose};
    foreach $letter (@LETTERS){
	InitializeLetter( $self, $letter ) if !defined $alphabet{$letter};
	
	foreach $word ( keys %{$alphabet{$letter}} ){
	    $val = $alphabet{$letter}{$word};
	    while( $val =~ m/<Sense>.*?<FIELD>(.*?)<\/FIELD>.*?<DEF>(.*?)<\/DEF>.*?<\/Sense>/ig ){
		push( @{$subjectCodes{$1}{$word}}, $2 );
	    }
	}
    }
    warn "END: Initializing the subject codes from the dictionary.\n" if $self->{verbose};
}

sub _initPosCodes( $ )
{
    my( $self ) = @_;
    my( $word, $val, $letter );
    
    # It is not defined so we need to load it and the whole hash while we are at it
    # Go through all the alphabet and load it
    warn "START: Initializing the POS codes from the dictionary ...\n" if $self->{verbose};
    foreach $letter (@LETTERS){
	InitializeLetter( $self, $letter ) if !defined $alphabet{$letter};
	
	foreach $word ( keys %{$alphabet{$letter}} ){
	    $val = $alphabet{$letter}{$word};
	    while( $val =~ m/<Head>.*?<POS>(.*?)<\/POS>.*?<\/Sense>/ig ){
		push( @{$posCodes{$1}}, $word );
	    }
	}
    }
    warn "END: Initializing the POS codes from the dictionary.\n" if $self->{verbose};
}


# Returns all the subject codes found in the dictionary
sub GetSubjectCodes( $ )
{
    my( $self ) = @_;
    
    # Check if we have it loaded or not
    if( $subjectCodesLoaded == 0){
	_initSubjectCodes( $self );
	
	$subjectCodesLoaded = 1;
    }
    
    return keys %subjectCodes;
}

# Returns all the senses that contain the specified word
sub HarvestSenses( $$ )
{
    my( $self, $word ) = @_;
    my( @defs ) = ();
    my( $sen ) = '';
    my( $letter ) = '';
    my( $w ) = '';
    my( $def ) = '';

    # First check in the cache
    if( exists $wsMap{$word} ){
	foreach $w (@{$wsMap{$word}}){
	    foreach $sen (GetSenses($self, $w)){
		$def = $sen->{def};
		push( @defs, $def ) if( $def =~ /\b$word\b/ );
	    }
	}
    }
    else{ # Word not in cache
	foreach $letter (@LETTERS){ # Have to look through entire dictionary
	    InitializeLetter( $self, $letter ) if !defined $alphabet{$letter};
	
	    foreach $w ( keys %{$alphabet{$letter}} ){
		foreach $sen (GetSenses($self, $w)){
		    $def = $sen->{def};
		    
		    next if !defined $def; # Undefined definition for this sense

		    if( $def =~ /\b$word\b/ ){
			push( @defs, $def );
			push( @{$wsMap{$word}}, $w ); # Cache it (the word)
		    }
		}
	    }
	}
    }
    
    @{$wsMap{$word}} = () if @defs == 0; # Nothing found for this word

    return @defs; # Done, return the senses
}

# Returns all the words that have one or more senses that contain the specified word
sub HarvestSensesByWords( $$ )
{
    my( $self, $word ) = @_;
    my( @words ) = ();
    my( $sen ) = '';
    my( $letter ) = '';
    my( $w ) = '';
    my( $def ) = '';	
    
    # Word not in cache
    if( !exists $wsMap{$word} ){
	foreach $letter (@LETTERS){ # Have to look through entire dictionary
	    InitializeLetter( $self, $letter ) if !defined $alphabet{$letter};
	
	    foreach $w ( keys %{$alphabet{$letter}} ){
		foreach $sen (GetSenses($self, $w)){
		    $def = $sen->{def};
		    
		    next if !defined $def; # Undefined definition for this sense
		    
		    push( @{$wsMap{$word}}, $w ) if( $def =~ /\b$word\b/ ); # Cache it (the word)
		}
	    }
	}
    }

    @{$wsMap{$word}} = () if @{$wsMap{$word}} == 0; # Nothing found for this word
    
    return @{$wsMap{$word}}; # Done, return the senses
}

# Returns all the words with a given subject code.
# NOTE: This operation is slow ONLY the first time it is used since the entire
# dictionary will have to be loaded. After that it becomes very fast
sub QuerySubjectCode( $$ )
{
    # Get the function arguments
    my( $self, $sc ) = @_;
    my( @allWords ) = ();

    # Check if we have it loaded or not
    if( $subjectCodesLoaded == 0){
	_initSubjectCodes( $self );	
	$subjectCodesLoaded = 1;
    }
    
    if( defined $subjectCodes{$sc} ){
	@allWords = keys %{$subjectCodes{$sc}};

    }

    return @allWords; # Return a reference to the array of all the words  
}

sub QuerySubjectCodeByWord( $$$$ )
{
    my( $self, $word, $sc, $incTags ) = @_;
    my( @senses ) = (); # The senses of this word as an array
    my( $sense ) = '';
    
    # Check for valid argument values
    return undef if !defined $word;
    return undef if !defined $sc;
    $incTags = '0' if !defined $incTags;

    # Check if we have it loaded or not
    if( $subjectCodesLoaded == 0){
	_initSubjectCodes( $self );
	
	$subjectCodesLoaded = 1;
    }
    
    
    if( $incTags eq '0' ){ # User does not want the tags so strip them out
	warn "QuerySubjectCodeByWord: Eliminating SGML Tags from word senses\n" if $self->{verbose};
	foreach $sense (@{$subjectCodes{$sc}{$word}}){
	    # Now we will strip the SGML tags from the sense
	    $sense =~ s/<REFSENSENUM>.+?<\/REFSENSENUM>/ /g;
	    $sense =~ s/<\/?\w+?>/ /g;
	    $sense =~ s/ +/ /g; # Substitute multiple spaces with one space
	    push( @senses, $sense );
	}
    }
    else{
	@senses = @{$subjectCodes{$sc}{$word}};
    }
    
    return @senses;
}

# Finds all the words with the same subject code as the given word
# Returns a HASH as follows:
# KEY  : subject code
# VALUE: array of all the words that belong to subject code
sub QueryRelated( $$ )
{
    my( $self, $word ) = @_;
    my( %relatedWords ) = ();
    my( $sgml ) = QuerySGML( $self, $word );
    my( @allSubjectCodes ) = (); # All the subject codes that belong to this word

    return %relatedWords if !defined $sgml; # Word not found in the dictionary
    
    while( $sgml =~ /<field>(.+?)<\/field>/gi ){
	push( @allSubjectCodes, $1 );
    }   

    # Now we have all the subject codes for this word
    my( @arr ) = ();
    my( $sc );
    foreach $sc (@allSubjectCodes){
	@arr = &QuerySubjectCode($self, $sc);
	# In each of the subject codes there must be at least one word (the word iteself)
	push( @{$relatedWords{$sc}}, @arr ) if @arr > 0;
    }    
    
    return %relatedWords;
}

sub QueryByPos( $$ )
{
    my( $self, $pos ) = @_;
    
    # Check if we have it loaded or not
    if( $posCodesLoaded == 0){
	warn "START: Initializing the POS codes from the dictionary ...\n" if $self->{verbose};
	_initPosCodes( $self );
	warn "END: Initializing the POS codes from the dictionary.\n" if $self->{verbose};
	
	$posCodesLoaded = 1;
    }

    my( @allWords ) = ();
    @allWords = @{$posCodes{$pos}} if exists $posCodes{$pos};

    return @allWords;
}


sub GetPosCodes( $ )
{
    my( $self ) = @_;
    
    if( $posCodesLoaded == 0 ){
	_initPosCodes( $self );
	
	$posCodesLoaded = 1;
    }

    return keys %posCodes;
}



sub QuerySGML
{
    # Get the word being retrieved
    my ($self, $word, $pos) = @_;
    my $letter;
    my $sense;

    return undef if( !defined $word );

    # Grab the letter to determine which hash to look into
    if( $word =~ /^(.)/i )
    {
	# Grab the letter and capitalize it
	$letter = $1;
	$letter =~ tr/a-z/A-Z/;
    }

    if( !defined($alphabet{$letter}) )
    {
	InitializeLetter($self, $letter);
    }
	
    my $sgmlData = $alphabet{$letter}{$word};
    
    # Check if a POS was passed as a parameter
    #
    if( defined($pos) )
    {
	# We need to match a specific part-of-speech only
	if( defined( $sgmlData ) )
	{
	    # Split this data into any different senses
	    my @senses = split /[\n]/, $sgmlData;
	    undef $sgmlData;
			
	    # Examine each one, looking for the POS asked about
	    #
	    for $sense (@senses)
	    {
		if( $sense =~ /<POS>$pos<\/POS>/i )
		{
		    # Now just set $sgmlData to this string and it will be
		    #	returned properly
		    $sgmlData = $sense;
		}
	    }
	}
    }
    
    # Now return the word asked for
    return $sgmlData;
}


# This will return the <HEAD> SGML data
#
sub GetHead
{
    my ($self, $word, $pos) = @_;
    my $sgmlData = QuerySGML($self, $word, $pos);
    my $sgmlHead;
    my @headData;
    my $headHash;
    my $entry;
    my $terminal;
	
    if( defined($sgmlData) )
    {
	my @senses = split /[\n]/, $sgmlData;
	
	for $entry (@senses)
	{
	    if( $entry =~ /(<HEAD>.*?<\/HEAD>)/i )
	    {
		$sgmlHead = $1;
		
		# Reset our reference hash
		$headHash = {};
				
		my @terminals = ('hwd','hyphenation','pos','gram','pron','littlewords','homnum','freq');
		
		for $terminal (@terminals)
		{
		    $headHash->{$terminal} = $1 if( $sgmlHead =~ /<$terminal>(.*?)<\/$terminal>/i );
		}
		
		push @headData, $headHash;
	    }
	}
    }
    
    return @headData;
}

sub GetSenses
{
    my ($self, $word, $pos ) = @_;
    my $sgmlData = QuerySGML($self, $word);;
    my $sgmlSense;
    my @senseData;
    my $senseHash;
    my $entry;
    my $terminal;
    
    if( defined($sgmlData) ){
	# Split this on the '\n' character
	my @senses = split /[\n]/, $sgmlData;
	
	for $entry (@senses){
	    my $entryPOS = $1 if ($entry =~ /<POS>(.*?)<\/POS>/i );
	    $entryPOS = '' if !defined $entryPOS;
	    
	    while( $entry =~ /(<Sense>.*?<\/Sense>)/ig )
	    {
		$senseHash = {};
		
		$sgmlSense = $1;
		
		# The following SGML tags are absolute in the sense they can't 
		#	have any sub-tags within them.
		#	
		#$senseHash->{sgml} = $sgmlSense;
		my @terminals = ('field','lexunit','activ','littlewords','lexvar','roman','colloexa','inflectype');
		
		for $terminal (@terminals){
		    $senseHash->{$terminal} = $1 if( $sgmlSense =~ /<$terminal>(.*?)<\/$terminal>/i );
		}
		
		$senseHash->{entry_pos} = $entryPOS;
		
		# The tags below are more complicated and may contain different
		#	sub-tags within them.
		#
		if( $sgmlSense =~ /<DEF>(.*?)<\/DEF>/i )
		{
		    # Process the defenition information by removing the 
		    #	extraneous tags (if any)
		    my $defInfo = $1;
		    $defInfo =~ s/<.*?>//g;
		    
		    $senseHash->{def} = $defInfo;
		}
		
		# There can be multiple examples so make sure to get all of them
		my $examples;
		while( $sgmlSense =~ /<EXAMPLE>(.*?)<\/EXAMPLE>/g )
		{
		    if( defined($examples) ) {
			$examples .= " | $1";
		    } else {
			$examples = "$1";
		    }
		}
		$senseHash->{example} = $examples if (defined($examples));
		
		if( (defined( $pos ) && ($pos eq $entryPOS)) || !defined( $pos ))
		{
		    push @senseData, $senseHash;
		}
	    }
	}
    }
    
    return @senseData;
}

sub GetSensesMultiple( $@ )
{
    my( $self, @allWords ) = @_;
    my( %defns ) = (); # The definitions of each word
    my( $word );

    foreach $word (@allWords){
	my @senses = &GetSenses( $self, $word );
	$defns{$word} = \@senses;
    }
    
    return %defns;
}


# Returns an associative array of all the subjects codes for a given word
sub GetSubjectCodesForWord( $$ )
{
    my( $self, $word ) = @_;
    my( %codes ) = ();

    return () if !defined $word || $word eq '';

    my( $sgml ) = QuerySGML( $self, $word );
    
    return () if !defined $sgml || $sgml eq '';
    
    while( $sgml =~ /<FIELD>(.+?)<\/FIELD>/gi ){
		$codes{$1} = undef;
    }

    return keys %codes;
}


# Returns an associative array of all the activ codes for a given word
sub GetActivCodesForWord( $$ )
{
    my( $self, $word ) = @_;
    my( %codes ) = ();
	my $code;

    return () if !defined $word || $word eq '';

    my( $sgml ) = QuerySGML( $self, $word );
    
    return () if !defined $sgml || $sgml eq '';
    
    while( $sgml =~ /<ACTIV>(.+?)<\/ACTIV>/gi ){
		$code = $1;
		$code =~ s/[#]//;
		$codes{$code} = undef;
    }

    return keys %codes;
}


# module must return true
1;
__END__

#################
# Documentation #
#################

=head1 NAME

LDOCE::Interface - A perl module to access the LDOCE3 MRD dictionary

=head1 SYNOPSIS

  use LDOCE::Interface;

  $ldoce = LDOCE::Interface->new;	
  $sgmlData = $ldoce->QuerySGML("apple");
  @senses = $ldoce->GetSenses("bat", 'v');

  # Alternative specifying custom location, initialization at start,
  # 	and verbose output
  $ldoce = LDOCE::Interface->new("/home/cs/tpederse/CS8761/LDOCE3/A-Z/sgml", 1, 1);

  $sgmlData = $ldoce->QuerySGML("apple");
  @headData = $ldoce->GetHead("bat", 'v');

  # Print out the head data
  for $href ( @head )
  {
    print "Head:\n";
    for $item (keys %$href)
    {
      print "\t$item -> $href->{$item}\n";
    }
    print "\n\n";
  }

=head1 REQUIRES

Perl 5.004, Exporter

=head1 EXPORTS

Nothing

=head1 DESCRIPTION

LDOCE::Interface provides an interface to the SGML version of the LDOCE
dictionary. This interface provides methods for grabbing information about
not only specific words, but also parts-of-speech and subject codes.

=head1 METHODS

=over 4

=item Default creation

$ldoce = LDOCE::Interface->new;

Creates and returns a handle by using the $LDOCEHOME environment variable or
else the module will default to a location specified in the module code. No
initialization is performed when created in this fashion. As different actions
are performed, the data will be initialized as needed.

=item Creation with custom parameters

$ldoce = LDOCE::Interface->new($path, $initialize, $verbose)

Creates a handle using the path specified. If $initialize is 1 then all the 
SGML data is processed when the module is created, resulting in very fast 
access later. If $verbose is 1, then the module will print more descriptive 
output to STDOUT.

=item Get the raw SGML data for a word

$sgmlData = $ldoce->QuerySGML($word [, $pos]); 

This returns the raw sgml data associated with $word, optionally restricted to the part-of-speech specified by $pos. If $pos is not specified, then the data for all available parts-of-speech is returned. The data in either case is returned as a simple string variable. In the event that multiple entries are present, they will be delimited by the newline ('\n') character.

=item Get the HEAD information for a word

@head = $ldoce->GetHead($word [, $pos]); 

This returns an array of hash references containing the data between the <HEAD> and </HEAD> tags for the word specified by $word. $pos can be optionally specified to only retrieve the data for a particular POS. The keys used correspond to the tags specified by the LDOCE dictionary. If a particular tag was not found in the sgml data, then it will not be set, so the resulting value for that key in the hash will be undefined.

=item Get the SENSE information for a word

@senses = $ldoce->GetSenses($word [, $pos]);

This returns an array of hash references containing the data for each "sense" present for a word. If $pos is specified then only those senses that are from a word patching that POS are returned. The keys used correspond to the tags present in the SGML data for a word. In the event that there is more than one entry for $word, each sense also contains a 'entry_pos' key to differeniate which entry it came from.

=item Get the SENSES of a list of words

%senses = $ldoce->GetSensesMultiple( @list_of_words );

This returns a hash of an array of hash references with the key being the word, containing 
the data for each "sense" present for a word. This function is basically a wrapper for the
function 

    $ldoce->getSenses( $word [, $pos] )

The data returned by the getSenses() function is put into a hash with the key being the 
word that the senses pertain to.
    Example: Print the all the senses of a list of words

    my( %defns ) = $ldoce->GetSensesMultiple( @words );
    my( @senses );
    my( $counter );
    foreach $word (keys %defns){
	$counter = 1;
	print "--- START WORD SENSE: $word\n";
	@senses = @{$defns{$word}};

        # Make sure that the word appears in the dictionary
        print "Word '${word}' not in dictionary\n" if @senses == 0;
	
	foreach $sense (@senses){
	    print "** SENSE $counter:\n";
	    while( ($key, $val) = each %$sense ){
		print "\t$key => $val\n";
	    }
	    $counter ++;
	}
	print "--- END WORD SENSE: $word\n\n";
    }    

=item Query words by Part Of Speech

@words = $LDOCE3->QueryByPos( PART_OF_SPEECH );

Returns all the words that belong to PART_OF_SPEECH as an associative array

Example: Get all the nouns

@words = $LDOCE3->QueryByPos( 'n' );

=item Get all available Part Of Speech codes

@pos = $LDOCE3->GetPosCodes();

Returns all the Part Of Speech codes as an associative array

=item Get all available subject codes

=item $ldoce->GetSubjectCodes();

Returns all the subject codes in the dictionary as an associative array.
    Example: @allCodes = $ldoce->GetSubjectCodes();

=item Query by subject code

@allWords = $LDOCE3->QuerySubjectCode( $subject_code )

$subject_code :: The subject code to query for. Case sensitive.

Returns all the words that belong to the subject code $subject_code. If the 
dictionary has not been loaded before this function has been called the first
time then the first call will be slow. This is because for this operation to 
be complete the dictionary will have to get loaded. This only happend on the 
first call.
    Example:
        @allWords = $LDOCE2->QuerySubjectCode( 'SCC' );

=item Query a by word with a specific subject code

    @senses = QuerySubjectCodeByWord( $word, $subject_code, $incTags );

       $word :: The word to query for
       $subject_code :: The subject code within the senses of this word. Case sensitive.
       $incTags :: Include SGML tags that are part of a sense

Returns all the senses of $word that belong to subject code $subject_code as
an associative array.
Some senses include SGML tags to refer to other words and to highlight certain
features of the sense. The $incTags flag is used to keep those tags or to remove
them from the returned senses. The tags will be removed if the value of $incTags=0
and kept otherwise. By default the tags are removed from the senses.
    Example:
        @allSenses = $LDOCE3->QuerySubjectCodeByWord( 'story', 'A' );

=item Get all related words

    %relatedWords = $ldoce->QueryRelated( $word );

This function will return all the words that are related to the given word. This
looks at the subject codes for the given word and then returns all the words that belong
to those subject codes. The return structure is a hash with the following form:
    # KEY  : subject code
    # VALUE: array of all the words that belong to subject code
In the case that the word is not in the dictionary then an empty hash is returned. Test this
using if( !%hash ), see example below for more details.

Example: Print all words related to 'canine'
    my( %result ) = $ldoce->QueryRelated( 'canine' );
    
    print "Word 'canine' not found in the dictionary\n" if !%result;

    my( @allWords );
    foreach $sc (keys %result){
	print "Subject Code: $sc\n";	
	print "\t".join( "\n\t", @{$result{$sc}} )."\n";
    }

=item Get all the subject codes of a given word

    @allCodes = $ldoce->GetSubjectCodesForWord( $$ );

This function will return an associative array of all the subject codes
of the given word. It can be considered as a wrapper for the function GetSenses($word)
but it returns only the subject code for the word

Example: Print all the subject codes of word 'decimal'
    my( @subjectCodes ) = $ldoce->GetSubjectCodesForWord( 'decimal' );
    foreach $code (@subjectCodes){
        print "$code \n";
    }

=item Get all the senses that contain a given word

    @allSenses = $ldoce->HarvestSenses( $$ );

This function will return an associative array that contains all the senses that contain
the requested word. The argument to the function ($word) is case sensitive

Example: Print all the senses that contain the word 'canine'
    my( @allSenses ) = $ldoce->HarvestSenses( 'canine' );
    foreach $sense (@allSenses){
        print "$sense \n";
    }

=item Get all the words that have senses that contain a given word

    @allWords = $ldoce->HarvestSensesByWords( $$ );

This function will returns all the words that have one or more senses that contain the specified word. 
The return is an associative array. The argument to the function ($word) is case sensitive

Example: Print all the words that have one or more senses that contain the word 'canine'
   my( @allWords ) = $ldoce->HarvestSensesByWords( 'canine' );
   foreach $word (@allWords){
      print "$word \n";
   }


=back

=head1 NOTES

Requires access to a directory containing the LDOCE sgml files (A.SGM, B.SGM, 
etc...). The case of the letters is not important, but they must be named with
the letter followed by ".SGM". This should be standard with the distribution of
the dictionary.

=head1 COPYRIGHT

Copyright 2002 Amine Abou-Rjeili and Sam Storie.  All rights reserved.

This module is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.

=head1 SEE ALSO

perl(1)

=cut

