# -*- perl -*-
#
# Package to fetch raw text using Altavista
#
# Run 'perldoc' on this file to produce documentation
#
# This module is free software; you can redistribute it and/or modify
# it under the same terms as Perl itself.
#
#
# Created for CS 8761 - Final Project
# A module to fetch raw text from Altavista's search engine
#
# Due: 11/25/2002
# Team: Alianza Lima
# Team Members:
# 	Abou-Rjeili, Amine
#	Goopy, Suchitra	
#	Purandare, Amruta	
#	Storie, Sam
#	

package WebReader::Interface;

use strict;
use Exporter;
use FileHandle;
use LWP::UserAgent;

##############################
# Environment/Initialization #
##############################

BEGIN {
    use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
    # List of classes from which we are inheriting methods
    @ISA = qw(Exporter);
    # Automatically loads these function names to be used without qualification
    @EXPORT = qw();
    # Allows these functions to be used without qualification
    @EXPORT_OK = qw();
    $VERSION = do { my @r=(q$Revision: 0.01 $=~/\d+/g); sprintf "%d."."%02d"x$#r,@r };
}

#############################
# Private Package Variables #
#############################

my $VERSION = '0.01';
my $baseDomain = 'http://www.altavista.com';
my $baseUrl = "$baseDomain/news/search?nc=0&nr=0&ns=0&nd=0&avkw=qtrp";
my $numRecords = 10;


# A variable to be our "user agent" when querying
my $ua;


END { } # module clean-up code here (global destructor)

###############
# Subroutines #
###############

sub new
{
	# The first argument is the class itself
	my $class = shift;
	my $self = {};
	
	# Make $self a reference to this class
	bless $self, $class;
	
	# Check for a verbose flag
	$self->{verbose} = @_ ? shift : 0;
	
	# Initialize the $ua variable
	$ua = LWP::UserAgent->new;
	$ua->agent("Mozilla/1.0");

	print "Module has been initialized.\n" if ($self->{verbose});

	return $self;
}


# This creates and issues a query to groups.google.com based on a
# 	set of query words. The result returned is an array of links
# 	to the plain-text newsgroup posting deemed relevant by google.
#
sub QuerySearchEngine
{
	my ($self, $query, $minRecords) = @_;
	my @urls;
	my $results = 0;
	my $perPage = 20;
	my $content;

	# If minRecords wasn't defined default to a single query
	#
	$minRecords = defined($minRecords) ? $minRecords : $numRecords;
	
	$query =~ s/ /+/g;
	
	print "Here's the query: $query\n" if ($self->{verbose});

	# This will return results from Altavista
	#
	my $queryString = "$baseUrl&q=$query";

	$content = FetchUrl($self, $queryString);
	
	if( defined($content) )
	{
		while( $content =~ /<a href=....ck_sm=.+?onMouseOver="status='(.+?)'/ig )
		{
			# Push this onto the array of urls if it's not another search string
			push @urls, "$1" if ( $1 !~ /search/i );
		}
	}

	sleep(2);

	return @urls;
}


# This just fetches the raw data from a specific url
#
sub FetchUrl( $$ )
{
	my ($self, $url) = @_;
	my $content = "";
	# Set up the actual request
	my $req = HTTP::Request->new(GET => $url);

	# Now issue the request
	print "Fetching: $url\n" if ($self->{verbose});
	my $res = $ua->request($req);

	if( $res->is_success )
	{
		print "Query was successful\n" if ($self->{verbose});
		$content = $res->content;
	}

	return $content;
}


# Parses a body of text and returns only 
# 	the sentences
#
sub ExtractSentences
{
	my ($self, $text, $minLength, $query) = @_;
	my @sentences;
	my $sentence;
	my $word;
	my @words;
	
	# If minLength isn't defined then make it zero
	#
	$minLength = defined($minLength) ? $minLength : 0;
	
	# If query is defined then split it into its seperate words
	#
	if( defined( $query ) )
	{
		print "Only extracting sentences that contain query words.\n" if ($self->{verbose});
		@words = split /\s/, $query if (defined($query));
	}
	
	# Remove the various header fields from the newsgroup posting
	#
	$text =~ s/\w+?:\s.+?\n//g; 

	# Remove anything that's not interesting
	# 
	$text =~ s/[^a-zA-Z0-9.,'"\-;:!?]/ /g;

	while( $text =~ /(.+?[.|!|?|"])\W+/g )
	{
		if( length $1 > $minLength )
		{
			$sentence = $1;

			# Now clean up the sentence a bit
			#
			$sentence =~ s/^\s+|\s+$//g;
			$sentence =~ s/\s{2,}/ /g;

			# Now if query has been specified, then restrict this to only
			#	those sentences that contain a query word
			if( defined($query) )
			{
				for $word (@words)
				{
					if( $sentence =~ /$word/i )
					{
						push @sentences, $sentence;
					}
				}
			}
			else
			{
				# Always push the sentence on
				push @sentences, $sentence;
			}
		}
	}
	
	printf "Extracted %d sentences.\n", scalar(@sentences) if ($self->{verbose});
	return @sentences;
}


sub RemoveHTML( $$$ )
{
	my ($self, $data, $maxTime) = @_;
	my $string = "";
	my $startTime;
	my $timeCheck;
	
	# Just rip through $data and remove any HTML found
	#
	# This specific regex was found at 
	#	http://aspn.activestate.com/ASPN/Cookbook/Rx/Recipe/59820
	#
	#$data =~ s/<(?:[^>'"]*|(['"]).*?\1)*>//gs;

	$startTime = time;
	
	# Remove all newlins
	$data =~ s/[\r\n]/ /g;

	RIP: while( $data =~ />(.+?)</g )
	{
		$string .= $1;

		if ( (time - $startTime) > $maxTime )
		{
			print "Max process time for HTML removal reached\n" if ($self->{verbose});
			last RIP;
		}
	}	
		
	# Remove any multiple spaces
	$string =~ s/\s{2,}/ /g;

	return $string;
}


# This is a large wrapper function to encompass all the tasks required
#	to collect a set of unique sentences based on a set of query words
#
# All parameters are mandatory as well
#
sub HarvestSentences( $$$$$$$ )
{
	my ($self, $query, $maxNumber, $minLength, $relative, $numRequests, $sleepSeconds) = @_;
	my @sentences;
	my $sentence;
	my @urls;
	my $url;
	my %uniqueSentences;
	my @totalSentences;
	my $page;
	my $urlCount = 0;
	my $key;
	my $existing;
	
	@urls = QuerySearchEngine($self, $query, $numRequests);
	
	printf "Found %d distinct pages to examine.\n", scalar(@urls) if ($self->{verbose});
	
	for $url (@urls)
	{
		# Check if we have the required amount and if so just cycle through
		#	the rest of the urls
		#
		next if( ($maxNumber != 0) && ($#totalSentences >= $maxNumber - 1) );
		
		$page = FetchUrl($self, $url);
		next if ( $page eq "" );
		
		$page = RemoveHTML($self, $page, 20);

		$urlCount++;
			
		# Extract the sentences from the page now
		#
		if( $relative == 1 )
		{
			@sentences = ExtractSentences($self, $page, $minLength, $query);
		}
		else
		{
			@sentences = ExtractSentences($self, $page, $minLength);
		}
		
		# Add these to our total list of sentences only if they don't
		#	already exist in our list	
		#
		for $sentence (@sentences)
		{
			if( !defined( $uniqueSentences{$sentence} ) )
			{
				# Define this sentence now
				$uniqueSentences{$sentence} = 1;
				
				push @totalSentences, $sentence;
			}
		}
		
		printf "Total sentences: %d\n", scalar(@totalSentences) if ($self->{verbose});
		
		# Sleep to prevent "aggressive" access
		#
		sleep($sleepSeconds);
	}
	
	return @totalSentences;
}



1;
__END__


#################
# Documentation #
#################


=head1 NAME

WebReader::Interface - Perl extension for fetching sentences from Google's
newsgroup archive.

=head1 SYNOPSIS

 use WebReader::Interface;
 my $numRequests = 500;
 my $minLength = 50;
 my $sleepSeconds = 3;
 my $relative = 0;
 
 print "Query (0 to exit): ";
 $query = <>;
 chomp($query);
 
 if( $query ne '0' )
 {
 	print "How many queries to make? ";
 	$numRequests = <>;
 	chomp($numRequests);
 	
 	print "Only save query-word containing sentences (y/n)? ";
 	$answer = <>;
 	chomp($answer);
 	$relative = 1 if ($answer =~ /^y/i);
 	
 	$fetch = WebReader::Interface->new(1);	
 	$fetch->HarvestSentences($query, $minLength, $relative, $numRequests, $sleepSeconds);
 	
 	print "\n\n";
 }

=head1 REQUIRES

Perl 5.004, Exporter, LWP::UserAgent

=head1 EXPORTS

Nothing

=head1 DESCRIPTION

WebReader::Interface is a module that allows you to send a query to Google's archive of
newsgroups and retrieve distinct sentences. This works by sending HTML requests,
via the LWP module, and parsing the resulting pages for the actual pages. Then
requests are issued for those pages and the sentences are retrieved.

When this module sends a query to google it really returns twice the number of
records. This is because it asks google to rank the resulting pages by both a
rank (determined by google) and the date the posting was created. So even though
google limits it's search replies to 1000 pages, you can actually retrieve more
pages with this method.


=head1 METHODS

=over 4

=item Default creation

$fetch = WebReader::Interface->new();

This creates the module and returns a handle to it. It also initializes the LWP::UserAgent
module for use.

=item Verbose creation

$fetch = WebReader::Interface->new(1);

This creates the module and returns a handle to it. It also initializes the LWP::UserAgent
module for use. The optional parameter defines the verbosity of the module.

=item Fetching a URL

$pageContent = $fetch->FetchUrl($url);

This function will return the raw HTML for a specific page. The only parameter is the
address of the page to fetch. If there is a problem then this returns an undefined value.

=item Gathering sentences

@totalSentences = $fetch->HarvestSentences($query, $minLength, $relative, $numRequests, $sleepSeconds);

This function is the main access point for this module. It performs all the tasks needed
to collect a large number of sentences and returns the result as an array containing
the sentences. $query is the actual query to use. $minLength is a number that dictates
the minimum length a sentence must be for it to be included. If $relative is 1 then a 
sentence must contain a query word for it to be included. $numRequests is a numerical 
value specifying the number of pages to search for via google. If 25 is specified for
example, then this module will fetch 25 results from google using each method described
in the DESCRIPTION section. $sleepSeconds is a number to tell the module how ling to 
sleep for between each page it fetches. A value less than 5 is not recommended but 
see the NOTES section for more detail.

Currently this module also saves the sentences in a file named after the seach query. For
example, if you searched for 'project mayhem' the sentences retrieved would be saved in
'project_mayhem.txt'. Should the file already exist the new data is appended to it.


=item Extracting sentences

@sentences = ExtractSentences($page, $minLength [, $query]);

This function can be used to extract sentences from a string of text. It is designed 
to be used primarily by HarvestSentences(). In the example shown above, $page would
be the string of raw text (preferebly not raw HTML), $minLength would be a numerical
value dictating the minimum sentence length. The optional $query value can be
used to restrict the sentences returned to only those containing a query word.

=back

=head1 NOTES

Because this module sends queries to Google there is an issue on how frequently
those requests should be sent. To prevent "aggressive" access you may want to 
introduce a longer delay when using the HarvestSentences() function. This delay
determines how long to wait when asking Google for the final pages. There is
a default 5 second delay when the module sends subsequent search requests. These
two delays are different and only the prior delay is configurable as a parameter.


=head1 COPYRIGHT

Copyright 2002 Sam Storie.  All rights reserved.

This module is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.

=head1 SEE ALSO

perl(1)

=cut

